//====================================================================
//Project Lynxmotion Phoenix
//Description: Phoenix, control file.
//The control input subroutine for the phoenix software is placed in this file.
//Can be used with V2.0 and above
//Configuration version: V1.0
//Date: 03-10-2017
//Programmer: Frank Friedhof port to PS3 USB/BT, Jeroen Janssen (aka Xan)
//             Kurt Eckhardt (aka KurtE) - converted to c ported to Arduino...
//
//Hardware setup: PS3 version
// 
//NEW IN V1.0
//- First Release
//
//Walk method 1:
//- Left Stick    Walk/Strafe
//- Right Stick   Rotate
//
//Walk method 2:
//- Left Stick    Disable
//- Right Stick   Walk/Rotate
//
//
//PS3 CONTROLS:
//[Common Controls]
//- StartTurn     on/off the bot
//- L1            Toggle Shift mode
//- L2            Toggle Rotate mode
//- Circle        Toggle Single leg mode
//- Square        Toggle Balance mode
//- Triangle      Move body to 35 mm from the ground (walk pos) 
//and back to the ground
//- D-Pad up      Body up 10 mm
//- D-Pad down    Body down 10 mm
//- D-Pad left    decrease speed with 50mS
//- D-Pad right   increase speed with 50mS
//
// Optional: L3 button down, Left stick can adjust leg positions...
// or if OPT_SINGLELEG is not defined may try using Circle

//
//
//[Walk Controls]
//- select        Switch gaits
//- Left Stick    (Walk mode 1) Walk/Strafe
//                (Walk mode 2) Disable
//- Right Stick   (Walk mode 1) Rotate, 
//                (Walk mode 2) Walk/Rotate
//- R1            Toggle Double gait travel speed
//- R2            Toggle Double gait travel length
//
//[Shift Controls]
//- Left Stick    Shift body X/Z
//- Right Stick   Shift body Y and rotate body Y
//
//[Rotate Controls]
//- Left Stick    Rotate body X/Z
//- Right Stick   Rotate body Y
//
//[Single leg Controls]
//- select        Switch legs
//- Left Stick    Move Leg X/Z (relative)
//- Right Stick   Move Leg Y (absolute)
//- R2            Hold/release leg position
//
//[GP Player Controls]
//- select        Switch Sequences
//- R2            Start Sequence
//
//====================================================================
// [Include files]
#if ARDUINO>99
#include <Arduino.h> // Arduino 1.0
#else
#include <Wprogram.h> // Arduino 0022
#endif
//#include <PS2X_lib.h>

//#define USE_PS3BT

#ifdef USE_PS3BT
#include <PS3BT.h>
#include <usbhub.h>
#else
#include <PS3USB.h>
#endif

//[CONSTANTS]
#define WALKMODE          0
#define TRANSLATEMODE     1
#define ROTATEMODE        2
#define SINGLELEGMODE     3
#define GPPLAYERMODE      4


#define cTravelDeadZone 4      //The deadzone for the analog input from the remote
#define  MAXPS2ERRORCNT  5     // How many times through the loop will we go before shutting off robot?

#ifndef MAX_BODY_Y
#define MAX_BODY_Y 100
#endif

//=============================================================================
// Global - Local to this file only...
//=============================================================================
//PS2X ps2x; // create PS2 Controller Class
USB Usb;
#ifdef USE_PS3BT
BTD Btd(&Usb); // You have to create the Bluetooth Dongle instance like so
PS3BT ps3x(&Btd, 0x00, 0x11, 0x22, 0x98, 0x76, 0x54); // This will also store the bluetooth address - this can be obtained from the dongle when running the sketch
#else
PS3USB ps3x(&Usb); // This will just create the instance
#endif

// Satisfy the IDE, which needs to see the include statment in the ino too.
#ifdef dobogusinclude
#include <spi4teensy3.h>
#include <SPI.h>
#endif



// Define an instance of the Input Controller...
InputController  g_InputController;       // Our Input controller 


static short      g_BodyYOffset; 
static short      g_sPS3ErrorCnt;
static short       g_BodyYShift;
static byte        ControlMode;
static bool        DoubleHeightOn;
static bool        DoubleTravelOn;
static bool        WalkMethod;
byte            GPSeq;             //Number of the sequence
short              g_sGPSMController;    // What GPSM value have we calculated. 0xff - Not used yet

// some external or forward function references.
extern void PS3TurnRobotOff(void);

//==============================================================================
// This is The function that is called by the Main program to initialize
//the input controller, which in this case is the PS2 controller
//process any commands.
//==============================================================================

// If both PS2 and XBee are defined then we will become secondary to the xbee
void InputController::Init(void)
{
  int error;

  //error = ps3x.config_gamepad(57, 55, 56, 54);  // Setup gamepad (clock, command, attention, data) pins
  //error = ps3x.config_gamepad(PS2_CLK, PS2_CMD, PS2_SEL, PS2_DAT);  // Setup gamepad (clock, command, attention, data) pins
  if (Usb.Init() == -1) {
    DBGSerial.print(F("\r\nOSC did not start"));
    while (1); //halt
  }
  DBGSerial.println(F("\r\nPS3 USB/Bluetooth Library Started"));

#ifdef DBGSerial
	DBGSerial.print("PS3 Init: ");
	DBGSerial.println(error, DEC);
#endif
  g_BodyYOffset = 0;    
  g_BodyYShift = 0;
  g_sPS3ErrorCnt = 0;  // error count

  ControlMode = WALKMODE;
  DoubleHeightOn = false;
  DoubleTravelOn = false;
  WalkMethod = false;

  g_InControlState.SpeedControl = 100;    // Sort of migrate stuff in from Devon.
}

//==============================================================================
// This function is called by the main code to tell us when it is about to
// do a lot of bit-bang outputs and it would like us to minimize any interrupts
// that we do while it is active...
//==============================================================================
void InputController::AllowControllerInterrupts(boolean fAllow)
{
  // We don't need to do anything...
}

//==============================================================================
// This is The main code to input function to read inputs from the PS2 and then
//process any commands.
//==============================================================================
#ifdef OPT_DYNAMIC_ADJUST_LEGS  
boolean g_fDynamicLegXZLength = false;  // Has the user dynamically adjusted the Leg XZ init pos (width)
#endif

void InputController::ControlInput(void)
{
  boolean fAdjustLegPositions = false;
  // Then try to receive a packet of information from the PS2.
  // Then try to receive a packet of information from the PS2.
//  ps3x.read_gamepad();          //read controller and set large motor to spin at 'vibrate' speed
  Usb.Task();

    // Wish the library had a valid way to verify that the read_gamepad succeeded...
    //  Will hack for now
  if (ps3x.PS3Connected || ps3x.PS3NavigationConnected) {
#ifdef DBGSerial
#ifdef DEBUG_PS3_INPUT
	if (g_fDebugOutput) {
		DBGSerial.print("PS3 Input: ");
// 		DBGSerial.print(ps3x.ButtonDataByte(), HEX);
// 		DBGSerial.print(":");
		DBGSerial.print(ps3x.getAnalogHat(LeftHatX), DEC);
		DBGSerial.print(" ");
		DBGSerial.print(ps3x.getAnalogHat(LeftHatY), DEC);
		DBGSerial.print(" ");
		DBGSerial.print(ps3x.getAnalogHat(RightHatX), DEC);
		DBGSerial.print(" ");
		DBGSerial.println(ps3x.getAnalogHat(RightHatY), DEC);
	}
#endif
#endif

#ifdef OPT_DYNAMIC_ADJUST_LEGS  
    boolean fAdjustLegPositions = false;
    short sLegInitXZAdjust = 0;
    short sLegInitAngleAdjust = 0;
#endif
    // In an analog mode so should be OK...
    g_sPS3ErrorCnt = 0;    // clear out error count...

    if (ps3x.getButtonClick(START)) {// OK lets press start button
      if (g_InControlState.fRobotOn) {
        DBGSerial.println("Robot turned OFF.");
        PS3TurnRobotOff();
      } 
      else {
        //Turn on
        DBGSerial.println("Robot turned ON.");
        g_InControlState.fRobotOn = 1;
        fAdjustLegPositions = true;
      }
    }

    if (g_InControlState.fRobotOn) {
      // [SWITCH MODES]

      //Translate mode
      if (ps3x.getButtonClick(L1)) {// L1 Button Test
        MSound( 1, 50, 2000);  
        if (ControlMode != TRANSLATEMODE ) {
          ControlMode = TRANSLATEMODE;
          DBGSerial.println("Translate mode");
        }
        else {
#ifdef OPT_SINGLELEG
          if (g_InControlState.SelectedLeg==255) {
            ControlMode = WALKMODE;
            DBGSerial.println("Walk mode");
          }
          else
#endif
          {
            ControlMode = SINGLELEGMODE;
            DBGSerial.println("Single leg mode");
          }
        }
      }

      //Rotate mode
      if (ps3x.getButtonClick(L2)) {    // L2 Button Test
        MSound( 1, 50, 2000);
        if (ControlMode != ROTATEMODE) {
          ControlMode = ROTATEMODE;
          DBGSerial.println("Rotate mode");
        }
        else {
#ifdef OPT_SINGLELEG
          if (g_InControlState.SelectedLeg == 255) {
            ControlMode = WALKMODE;
            DBGSerial.println("Walk mode");
          }
          else
#endif
          {
            ControlMode = SINGLELEGMODE;
            DBGSerial.println("Single leg mode");
          }
        }
      }

      //Single leg mode fNO
#ifdef OPT_SINGLELEG
      if (ps3x.getButtonClick(CIRCLE)) {// O - Circle Button Test
        if (abs(g_InControlState.TravelLength.x)<cTravelDeadZone && abs(g_InControlState.TravelLength.z)<cTravelDeadZone 
          && abs(g_InControlState.TravelLength.y*2)<cTravelDeadZone )   {
          if (ControlMode != SINGLELEGMODE) {
            ControlMode = SINGLELEGMODE;
            DBGSerial.println("Single leg mode");
            if (g_InControlState.SelectedLeg == 255)  //Select leg if none is selected
              g_InControlState.SelectedLeg=cRF; //Startleg
          } 
          else {
            ControlMode = WALKMODE;
            DBGSerial.println("Walk mode");
            g_InControlState.SelectedLeg=255;
          }
        }
      }      
#endif
#ifdef OPT_GPPLAYER
      // GP Player Mode X
      if (ps3x.getButtonClick(CROSS)) { // X - Cross Button Test
        MSound(1, 50, 2000);  
        if (ControlMode != GPPLAYERMODE) {
          ControlMode = GPPLAYERMODE;
          GPSeq=0;
          DBGSerial.println("GP player mode");
        } 
        else {
          ControlMode = WALKMODE;
          DBGSerial.println("Walk mode");
        }
      }
#endif // OPT_GPPLAYER

      //[Common functions]
      //Switch Balance mode on/off 
      if (ps3x.getButtonClick(SQUARE)) { // Square Button Test
        g_InControlState.BalanceMode = !g_InControlState.BalanceMode;
        if (g_InControlState.BalanceMode) {
          MSound(1, 250, 1500);
          DBGSerial.println("Balance mode ON");
        } 
        else {
          MSound( 2, 100, 2000, 50, 4000);
          DBGSerial.println("Balance mode OFF");
        }
        DBGSerial.println("Beep :-)");
      }

      //Stand up, sit down  
      if (ps3x.getButtonClick(TRIANGLE)) { // Triangle - Button Test
        if (g_BodyYOffset>0) 
          g_BodyYOffset = 0;
        else
          g_BodyYOffset = 35;
        fAdjustLegPositions = true;
        DBGSerial.print("BodyYOffset : ");
        DBGSerial.println(g_BodyYOffset);
      }

      if (ps3x.getButtonClick(UP)) {// D-Up - Button Test
        g_BodyYOffset += 10;

        // And see if the legs should adjust...
        fAdjustLegPositions = true;
        if (g_BodyYOffset > MAX_BODY_Y)
          g_BodyYOffset = MAX_BODY_Y;
        DBGSerial.print("BodyYOffset : ");
        DBGSerial.println(g_BodyYOffset);
      }

      if (ps3x.getButtonClick(DOWN) && g_BodyYOffset) {// D-Down - Button Test
        if (g_BodyYOffset > 10)
          g_BodyYOffset -= 10;
        else
          g_BodyYOffset = 0;      // constrain don't go less than zero.

        // And see if the legs should adjust...
        fAdjustLegPositions = true;
        DBGSerial.print("BodyYOffset : ");
        DBGSerial.println(g_BodyYOffset);
      }

      if (ps3x.getButtonClick(RIGHT)) { // D-Right - Button Test
        if (g_InControlState.SpeedControl>0) {
          g_InControlState.SpeedControl = g_InControlState.SpeedControl - 50;
          MSound( 1, 50, 2000);  
        }
        DBGSerial.print("SpeedControl : ");
        DBGSerial.println(g_InControlState.SpeedControl);
      }

      if (ps3x.getButtonClick(LEFT)) { // D-Left - Button Test
        if (g_InControlState.SpeedControl<2000 ) {
          g_InControlState.SpeedControl = g_InControlState.SpeedControl + 50;
          MSound( 1, 50, 2000); 
        }
        DBGSerial.print("SpeedControl : ");
        DBGSerial.println(g_InControlState.SpeedControl);
      }
      
      // We are optionally going to allow the user to modify the Initial Leg positions, when they
      // press the L3 button.
      byte lx = ps3x.getAnalogHat(LeftHatX);
      byte ly = ps3x.getAnalogHat(LeftHatY);
#ifdef OPT_DYNAMIC_ADJUST_LEGS  
#ifdef OPT_SINGLELEG
      if (ps3x.getButtonPress(L3)) {    // L3 pressed, use this to modify leg positions.
#else
      if (ps3x.getButtonPress(CIRCLE)) {// O - Circle Button Test 
#endif      
        sLegInitXZAdjust = ((int)lx-128)/10;        // play with this.
        sLegInitAngleAdjust = ((int)ly-128)/8;
        lx = 0;
        ly = 0;
      }
#endif

      //[Walk functions]
      if (ControlMode == WALKMODE) {
        //Switch gates
        if (ps3x.getButtonPress(SELECT)            // Select Button Test
        && abs(g_InControlState.TravelLength.x)<cTravelDeadZone //No movement
        && abs(g_InControlState.TravelLength.z)<cTravelDeadZone 
          && abs(g_InControlState.TravelLength.y*2)<cTravelDeadZone  ) {
          g_InControlState.GaitType = g_InControlState.GaitType+1;                    // Go to the next gait...
          if (g_InControlState.GaitType < NUM_GAITS) {                 // Make sure we did not exceed number of gaits...
            MSound( 1, 50, 2000); 
          } 
          else {
            MSound(2, 50, 2000, 50, 2250); 
            g_InControlState.GaitType = 0;
          }
          GaitSelect();
        }

        //Double leg lift height
        if (ps3x.getButtonClick(R1)) { // R1 Button Test
          MSound( 1, 50, 2000); 
          DoubleHeightOn = !DoubleHeightOn;
          if (DoubleHeightOn)
            g_InControlState.LegLiftHeight = 80;
          else
            g_InControlState.LegLiftHeight = 50;
          DBGSerial.print("Gait level ");
          DBGSerial.println(g_InControlState.LegLiftHeight);
        }

        //Double Travel Length
        if (ps3x.getButtonPress(R2)) {// R2 Button Test
          MSound(1, 50, 2000); 
          DoubleTravelOn = !DoubleTravelOn;
          DBGSerial.print("Gait speed ");
          DBGSerial.println(DoubleTravelOn?"normal":"double");
        }

        // Switch between Walk method 1 && Walk method 2
        if (ps3x.getButtonPress(R3)) { // R3 Button Test
          MSound(1, 50, 2000); 
          WalkMethod = !WalkMethod;
          DBGSerial.print("Walk method ");
          DBGSerial.println(WalkMethod?"1":"2");
        }

        //Walking
        if (WalkMethod)  //(Walk Methode) 
          g_InControlState.TravelLength.z = (ps3x.getAnalogHat(RightHatY)-128); //Right Stick Up/Down  

        else {
          g_InControlState.TravelLength.x = -(lx - 128);
          g_InControlState.TravelLength.z = (ly - 128);
        }

        if (!DoubleTravelOn) {  //(Double travel length)
          g_InControlState.TravelLength.x = g_InControlState.TravelLength.x/2;
          g_InControlState.TravelLength.z = g_InControlState.TravelLength.z/2;
        }

        g_InControlState.TravelLength.y = -(ps3x.getAnalogHat(RightHatX) - 128)/4; //Right Stick Left/Right 
      }

      //[Translate functions]
      g_BodyYShift = 0;
      if (ControlMode == TRANSLATEMODE) {
        g_InControlState.BodyPos.x = (lx - 128)/2;
        g_InControlState.BodyPos.z = -(ly - 128)/3;
        g_InControlState.BodyRot1.y = (ps3x.getAnalogHat(RightHatX) - 128)*2;
        g_BodyYShift = (-(ps3x.getAnalogHat(RightHatY) - 128)/2);
      }

      //[Rotate functions]
      if (ControlMode == ROTATEMODE) {
        g_InControlState.BodyRot1.x = (ly - 128);
        g_InControlState.BodyRot1.y = (ps3x.getAnalogHat(RightHatX) - 128)*2;
        g_InControlState.BodyRot1.z = (lx - 128);
        g_BodyYShift = (-(ps3x.getAnalogHat(RightHatY) - 128)/2);
      }

      //[Single leg functions]
#ifdef OPT_SINGLELEG
      if (ControlMode == SINGLELEGMODE) {
        //Switch leg for single leg control
        if (ps3x.getButtonPress(SELECT)) { // Select Button Test
          MSound(1, 50, 2000); 
          if (g_InControlState.SelectedLeg<(CNT_LEGS-1))
            g_InControlState.SelectedLeg = g_InControlState.SelectedLeg+1;
          else
            g_InControlState.SelectedLeg=0;
        }

        g_InControlState.SLLeg.x= (lx - 128)/2; //Left Stick Right/Left
        g_InControlState.SLLeg.y= (ps3x.getAnalogHat(RightHatY) - 128)/10; //Right Stick Up/Down
        g_InControlState.SLLeg.z = (ly - 128)/2; //Left Stick Up/Down

        // Hold single leg in place
        if (ps3x.getButtonPress(R2)) { // R2 Button Test
          MSound(1, 50, 2000);  
          g_InControlState.fSLHold = !g_InControlState.fSLHold;
        }
      }
#endif
#ifdef OPT_GPPLAYER
      //[GPPlayer functions]
      if (ControlMode == GPPLAYERMODE) {

        // Lets try some speed control... Map all values if we have mapped some before
        // or start mapping if we exceed some minimum delta from center
        // Have to keep reminding myself that commander library already subtracted 128...
        if (g_ServoDriver.FIsGPSeqActive() ) {
          if ((g_sGPSMController != 32767)  
            || (ps3x.getAnalogHat(RightHatY) > (128+16)) || (ps3x.getAnalogHat(RightHatY) < (128-16)))
          {
            // We are in speed modify mode...
            short sNewGPSM = map(ps3x.getAnalogHat(RightHatY), 0, 255, -200, 200);
            if (sNewGPSM != g_sGPSMController) {
              g_sGPSMController = sNewGPSM;
              g_ServoDriver.GPSetSpeedMultiplyer(g_sGPSMController);
            }

          }
        }

        //Switch between sequences
        if (ps3x.getButtonClick(SELECT)) { // Select Button Test
          if (!g_ServoDriver.FIsGPSeqActive() ) {
            if (GPSeq < 5) {  //Max sequence
              MSound(1, 50, 1500);
              GPSeq = GPSeq+1;
            } 
            else {
              MSound(2, 50, 2000, 50, 2250);
              GPSeq=0;
            }
            DBGSerial.print("GP sequence ");
            DBGSerial.println(GPSeq, DEC);
          }
        }
        //Start Sequence
        if (ps3x.getButtonClick(R2))// R2 Button Test
          if (!g_ServoDriver.FIsGPSeqActive() ) {
            g_ServoDriver.GPStartSeq(GPSeq);
            g_sGPSMController = 32767;  // Say that we are not in Speed modify mode yet... valid ranges are 50-200 (both postive and negative... 
          }
          else {
            g_ServoDriver.GPStartSeq(0xff);    // tell the GP system to abort if possible...
            MSound (2, 50, 2000, 50, 2000);
          }


      }
#endif // OPT_GPPLAYER

      //Calculate walking time delay
      g_InControlState.InputTimeDelay = 128 - max(max(abs(lx - 128), abs(ly - 128)), abs(ps3x.getAnalogHat(RightHatX) - 128));
    }

    //Calculate g_InControlState.BodyPos.y
    g_InControlState.BodyPos.y = min(max(g_BodyYOffset + g_BodyYShift,  0), MAX_BODY_Y);
    
#ifdef OPT_DYNAMIC_ADJUST_LEGS  
    if (sLegInitXZAdjust || sLegInitAngleAdjust) {
      // User asked for manual leg adjustment - only do when we have finished any previous adjustment

      if (!g_InControlState.ForceGaitStepCnt) {
        if (sLegInitXZAdjust)
          g_fDynamicLegXZLength = true;

        sLegInitXZAdjust += GetLegsXZLength();  // Add on current length to our adjustment...
        // Handle maybe change angles...
        if (sLegInitAngleAdjust) 
            RotateLegInitAngles(sLegInitAngleAdjust);
        // Give system time to process previous calls
        AdjustLegPositions(sLegInitXZAdjust);
      }
    }    
#endif
    
    if (fAdjustLegPositions)
      AdjustLegPositionsToBodyHeight();    // Put main workings into main program file
  } 
  else {
    // We may have lost the PS2... See what we can do to recover...
    if (g_sPS3ErrorCnt < MAXPS2ERRORCNT)
      g_sPS3ErrorCnt++;    // Increment the error count and if to many errors, turn off the robot.
    else if (g_InControlState.fRobotOn)
      PS3TurnRobotOff();
//    ps3x.reconfig_gamepad();
  }
}

//==============================================================================
// PS3TurnRobotOff - code used couple of places so save a little room...
//==============================================================================
void PS3TurnRobotOff(void)
{
  //Turn off
  g_InControlState.BodyPos.x = 0;
  g_InControlState.BodyPos.y = 0;
  g_InControlState.BodyPos.z = 0;
  g_InControlState.BodyRot1.x = 0;
  g_InControlState.BodyRot1.y = 0;
  g_InControlState.BodyRot1.z = 0;
  g_InControlState.TravelLength.x = 0;
  g_InControlState.TravelLength.z = 0;
  g_InControlState.TravelLength.y = 0;
  g_BodyYOffset = 0;
  g_BodyYShift = 0;
#ifdef OPT_SINGLELEG
  g_InControlState.SelectedLeg = 255;
#endif  
  g_InControlState.fRobotOn = 0;
  AdjustLegPositionsToBodyHeight();    // Put main workings into main program file
}




